@@ -52,7 +52,7 @@ static guint    pixbuf_loader_signals[LAST_SIGNAL] = { 0 };
 
 /* Internal data */
 
-#define LOADER_HEADER_SIZE 128
+#define LOADER_HEADER_SIZE 1024
 
 typedef struct
 {
@@ -485,7 +485,7 @@ gdk_pixbuf_loader_write (GdkPixbufLoader *loader,
         /* we expect it's not to be closed */
         g_return_val_if_fail (priv->closed == FALSE, FALSE);
   
-        if (priv->image_module == NULL)
+        if (count > 0 && priv->image_module == NULL)
                 {
                         gint eaten;
       
@@ -547,6 +547,12 @@ gdk_pixbuf_loader_new (void)
  * that can't be reliably identified by looking at the data, or if
  * the user manually forces a specific type.
  *
+ * The list of supported image formats depends on what image loaders
+ * are installed, but typically "png", "jpeg", "gif", "tiff" and 
+ * "xpm" are among the supported formats. To obtain the full list of
+ * supported image formats, call gdk_pixbuf_format_get_name() on each 
+ * of the #GdkPixbufFormat structs returned by gdk_pixbuf_get_formats().
+ *
  * Return value: A newly-created pixbuf loader.
  **/
 GdkPixbufLoader *
@@ -584,6 +590,13 @@ gdk_pixbuf_loader_new_with_type (const char *image_type,
  * that can't be reliably identified by looking at the data, or if
  * the user manually forces a specific mime type.
  *
+ * The list of supported mime types depends on what image loaders
+ * are installed, but typically "image/png", "image/jpeg", "image/gif", 
+ * "image/tiff" and "image/x-xpixmap" are among the supported mime types. 
+ * To obtain the full list of supported mime types, call 
+ * gdk_pixbuf_format_get_mime_types() on each of the #GdkPixbufFormat 
+ * structs returned by gdk_pixbuf_get_formats().
+ *
  * Return value: A newly-created pixbuf loader.
  * Since: 2.4
  **/
@@ -727,7 +740,9 @@ gdk_pixbuf_loader_close (GdkPixbufLoader *loader,
         /* we expect it's not closed */
         g_return_val_if_fail (priv->closed == FALSE, TRUE);
   
-        /* We have less the 128 bytes in the image.  Flush it, and keep going. */
+        /* We have less the LOADER_HEADER_SIZE bytes in the image.  
+         * Flush it, and keep going. 
+         */
         if (priv->image_module == NULL)
                 {
                         GError *tmp = NULL;
