@@ -24,6 +24,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include <math.h>
 #include <stdlib.h>
 #include <string.h>
@@ -87,7 +88,8 @@ gdk_font_hash_lookup (GdkFontType type, const gchar *font_name)
 }
 
 GdkFont*
-gdk_font_from_description (PangoFontDescription *desc)
+gdk_font_from_description_for_display (GdkDisplay           *display,
+				       PangoFontDescription *desc)
 {
   GdkFont *font;
   GdkFontPrivateFB *private;
@@ -140,7 +142,7 @@ gdk_font_from_description (PangoFontDescription *desc)
   private->pango_font = pango_font;
   
   g_free (lang);
-  g_object_unref (G_OBJECT (context));
+  g_object_unref (context);
 
   font->ascent = PANGO_PIXELS (pango_font_metrics_get_ascent (metrics));
   font->descent = PANGO_PIXELS (pango_font_metrics_get_descent (metrics));
@@ -154,7 +156,8 @@ gdk_font_from_description (PangoFontDescription *desc)
 
 
 GdkFont*
-gdk_font_load (const gchar *font_name)
+gdk_font_load_for_display (GdkDisplay  *display,
+			   const gchar *font_name)
 {
   GdkFont *font;
   GdkFontPrivateFB *private;
@@ -243,7 +246,7 @@ _gdk_font_destroy (GdkFont *font)
   GdkFontPrivateFB *private = (GdkFontPrivateFB *)font;
   gdk_font_hash_remove (font->type, font);
 
-  g_object_unref (G_OBJECT (private->pango_font));
+  g_object_unref (private->pango_font);
   g_free (private->name);
   g_free (font);
 }
